@@ -1,5 +1,5 @@
-/* Copyright (C) 1995, 1996, 1998, 1999, 2000, 2001, 2004, 2006, 2008,
- *   2009, 2010, 2011, 2012, 2013, 2014, 2017 Free Software Foundation, Inc.
+/* Copyright (C) 1995, 1996, 1998-2001, 2004, 2006, 2008-2018
+ *   Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -146,7 +146,7 @@ SCM_DEFINE (scm_sys_package_data_dir, "%package-data-dir", 0, 0, 0,
 	    "@samp{/usr/local/share/guile}.")
 #define FUNC_NAME s_scm_sys_package_data_dir
 {
-  return scm_from_locale_string (SCM_PKGDATA_DIR);
+  return scm_from_utf8_string (SCM_PKGDATA_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_PKGDATA_DIR */
@@ -158,7 +158,7 @@ SCM_DEFINE (scm_sys_library_dir, "%library-dir", 0,0,0,
 	    "E.g., may return \"/usr/share/guile/1.3.5\".")
 #define FUNC_NAME s_scm_sys_library_dir
 {
-  return scm_from_locale_string (SCM_LIBRARY_DIR);
+  return scm_from_utf8_string (SCM_LIBRARY_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_LIBRARY_DIR */
@@ -184,7 +184,7 @@ SCM_DEFINE (scm_sys_global_site_dir, "%global-site-dir", 0,0,0,
 	    "E.g., may return \"/usr/share/guile/site\".")
 #define FUNC_NAME s_scm_sys_global_site_dir
 {
-  return scm_from_locale_string (SCM_GLOBAL_SITE_DIR);
+  return scm_from_utf8_string (SCM_GLOBAL_SITE_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_GLOBAL_SITE_DIR */
@@ -197,7 +197,7 @@ SCM_DEFINE (scm_sys_site_ccache_dir, "%site-ccache-dir", 0,0,0,
 	    "E.g., may return \"/usr/lib/guile/" SCM_EFFECTIVE_VERSION "/site-ccache\".")
 #define FUNC_NAME s_scm_sys_site_ccache_dir
 {
-  return scm_from_locale_string (SCM_SITE_CCACHE_DIR);
+  return scm_from_utf8_string (SCM_SITE_CCACHE_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_SITE_CCACHE_DIR */
@@ -302,6 +302,9 @@ scm_i_mirror_backslashes (char *path)
 
       while (*p)
 	{
+          /* FIXME: When the locale encoding is Shift_JIS, backslash '\'
+             has a multibyte representation, so this code will
+             misbehave.  */
 	  if (*p == '\\')
 	    *p = '/';
 	  p++;
@@ -331,10 +334,10 @@ scm_init_load_path ()
   else if (env)
     path = scm_parse_path (scm_from_locale_string (env), path);
   else
-    path = scm_list_4 (scm_from_locale_string (SCM_LIBRARY_DIR),
-                       scm_from_locale_string (SCM_SITE_DIR),
-                       scm_from_locale_string (SCM_GLOBAL_SITE_DIR),
-                       scm_from_locale_string (SCM_PKGDATA_DIR));
+    path = scm_list_4 (scm_from_utf8_string (SCM_LIBRARY_DIR),
+                       scm_from_utf8_string (SCM_SITE_DIR),
+                       scm_from_utf8_string (SCM_GLOBAL_SITE_DIR),
+                       scm_from_utf8_string (SCM_PKGDATA_DIR));
 
   env = scm_i_mirror_backslashes (getenv ("GUILE_SYSTEM_COMPILED_PATH"));
   if (env && strcmp (env, "") == 0)
@@ -344,8 +347,8 @@ scm_init_load_path ()
     cpath = scm_parse_path (scm_from_locale_string (env), cpath);
   else
     {
-      cpath = scm_list_2 (scm_from_locale_string (SCM_CCACHE_DIR),
-                          scm_from_locale_string (SCM_SITE_CCACHE_DIR));
+      cpath = scm_list_2 (scm_from_utf8_string (SCM_CCACHE_DIR),
+                          scm_from_utf8_string (SCM_SITE_CCACHE_DIR));
     }
 
 #endif /* SCM_LIBRARY_DIR */
@@ -946,7 +949,7 @@ SCM_DEFINE (scm_search_path, "search-path", 2, 0, 1,
 	    {
 	      require_exts = SCM_CADR (rest);
 	      if (SCM_UNLIKELY (!scm_is_null (SCM_CDDR (rest))))
-		scm_wrong_num_args (scm_from_locale_string (FUNC_NAME));
+		scm_wrong_num_args (scm_from_utf8_string (FUNC_NAME));
 	    }
 	}
       else
@@ -1195,16 +1198,13 @@ SCM_DEFINE (scm_primitive_load_path, "primitive-load-path", 0, 0, 1,
                                *scm_loc_load_extensions, SCM_BOOL_F,
                                &stat_source);
 
-  if (scm_is_false (*scm_loc_fresh_auto_compile))
-    compiled_thunk = load_thunk_from_path (filename, full_filename,
-					   &stat_source,
-					   &found_stale_compiled_file);
-  else
-    compiled_thunk = SCM_BOOL_F;
+  compiled_thunk = load_thunk_from_path (filename, full_filename, &stat_source,
+                                         &found_stale_compiled_file);
 
   if (scm_is_false (compiled_thunk)
       && scm_is_true (full_filename)
       && scm_is_true (*scm_loc_compile_fallback_path)
+      && scm_is_false (*scm_loc_fresh_auto_compile)
       && scm_is_pair (*scm_loc_load_compiled_extensions)
       && scm_is_string (scm_car (*scm_loc_load_compiled_extensions)))
     {
@@ -1275,10 +1275,10 @@ scm_init_eval_in_scheme (void)
   int found_stale_eval_go = 0;
 
   eval_scm = search_path (*scm_loc_load_path,
-                          scm_from_locale_string ("ice-9/eval.scm"),
+                          scm_from_utf8_string ("ice-9/eval.scm"),
                           SCM_EOL, SCM_BOOL_F, &stat_source);
   eval_thunk =
-    load_thunk_from_path (scm_from_locale_string ("ice-9/eval.go"),
+    load_thunk_from_path (scm_from_utf8_string ("ice-9/eval.go"),
                           eval_scm, &stat_source, &found_stale_eval_go);
   
   if (scm_is_true (eval_thunk))
@@ -1336,13 +1336,13 @@ scm_init_load ()
   scm_loc_load_path = SCM_VARIABLE_LOC (scm_c_define ("%load-path", SCM_EOL));
   scm_loc_load_extensions
     = SCM_VARIABLE_LOC (scm_c_define ("%load-extensions",
-				      scm_list_2 (scm_from_locale_string (".scm"),
+				      scm_list_2 (scm_from_utf8_string (".scm"),
 						  scm_nullstr)));
   scm_loc_load_compiled_path
     = SCM_VARIABLE_LOC (scm_c_define ("%load-compiled-path", SCM_EOL));
   scm_loc_load_compiled_extensions
     = SCM_VARIABLE_LOC (scm_c_define ("%load-compiled-extensions",
-				      scm_list_1 (scm_from_locale_string (".go"))));
+				      scm_list_1 (scm_from_utf8_string (".go"))));
   scm_loc_load_hook = SCM_VARIABLE_LOC (scm_c_define ("%load-hook", SCM_BOOL_F));
 
   scm_loc_compile_fallback_path
