@@ -1,5 +1,5 @@
-/* Copyright (C) 1999, 2000, 2001, 2003, 2005, 2006, 2009, 2010,
- *    2012, 2013, 2014 Free Software Foundation, Inc.
+/* Copyright (C) 1999-2001, 2003, 2005, 2006, 2009, 2010, 2012-2014,
+ *   2017, 2018 Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -374,7 +374,10 @@ make_rstate (scm_t_rstate *state)
  * Scheme level interface.
  */
 
-SCM_GLOBAL_VARIABLE_INIT (scm_var_random_state, "*random-state*", scm_seed_to_random_state (scm_from_locale_string ("URL:http://stat.fsu.edu/~geo/diehard.html")));
+SCM_GLOBAL_VARIABLE_INIT (scm_var_random_state, "*random-state*",
+                          scm_seed_to_random_state
+                          (scm_from_utf8_string
+                           ("URL:http://stat.fsu.edu/~geo/diehard.html")));
 
 SCM_DEFINE (scm_random, "random", 1, 1, 0, 
             (SCM n, SCM state),
@@ -438,11 +441,33 @@ SCM_DEFINE (scm_seed_to_random_state, "seed->random-state", 1, 0, 0,
 #define FUNC_NAME s_scm_seed_to_random_state
 {
   SCM res;
+  char *c_str;
+  size_t len;
+
   if (SCM_NUMBERP (seed))
     seed = scm_number_to_string (seed, SCM_UNDEFINED);
   SCM_VALIDATE_STRING (1, seed);
-  res = make_rstate (scm_c_make_rstate (scm_i_string_chars (seed),
-					scm_i_string_length (seed)));
+
+  if (scm_i_is_narrow_string (seed))
+    /* This special case of a narrow string, where latin1 is used, is
+       for backward compatibility during the 2.2 stable series.  In
+       future major releases, we should use UTF-8 uniformly. */
+    c_str = scm_to_latin1_stringn (seed, &len);
+  else
+    c_str = scm_to_utf8_stringn (seed, &len);
+
+  /* 'scm_to_*_stringn' returns a 'size_t' for the length in bytes, but
+     'scm_c_make_rstate' accepts an 'int'.  Make sure the length fits in
+     an 'int'. */
+  if (len > INT_MAX)
+    {
+      free (c_str);
+      SCM_OUT_OF_RANGE (1, seed);
+    }
+
+  res = make_rstate (scm_c_make_rstate (c_str, len));
+  free (c_str);
+
   scm_remember_upto_here_1 (seed);
   return res;
   
